/*
 *  Copyright (C) 2006 Ludovic Jacomme (ludovic.jacomme@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>

#include "cap32.h"
#include "psp_kbd.h"
#include "psp_sdl.h"
#include "psp_menu.h"
#include "psp_menu_help.h"

# define MAX_HELP_LINE    4096 

  static char* psp_help[MAX_HELP_LINE];
  static int   psp_help_size    = -1;
  static int   psp_help_current = 0;
  static int   psp_help_line_by_page = 10;
  static int   psp_help_char_by_line = 40;

static void
psp_initialize_help(void)
{
  char  FileName[MAX_PATH+1];

  char  Buffer[512];
  char *Scan;
  FILE* FileDesc;

  /* Already done ? */
  if (psp_help_size > 0) return; 

  int font_width  = psp_sdl_get_ttf_width();
  int font_heigth = psp_sdl_get_ttf_height();

  psp_help_char_by_line = (PSP_SDL_SCREEN_WIDTH  - 2 * font_width ) / font_width;
  psp_help_line_by_page = (PSP_SDL_SCREEN_HEIGHT                  ) / font_heigth;

  strcpy(FileName, "./help.txt");
  FileDesc = fopen(FileName, "r");

  psp_help_current = 0;

  if (FileDesc == (FILE *)0 ) {
    psp_help[0] = strdup( "no help file found !");
    psp_help_size = 1;
    return;
  }

  psp_help_size = 0;
  while (fgets(Buffer,512, FileDesc) != (char *)0) {

    Scan = strchr(Buffer,'\n');
    if (Scan) *Scan = '\0';
    /* For this #@$% of windows ! */
    Scan = strchr(Buffer,'\r');
    if (Scan) *Scan = '\0';

    psp_help[psp_help_size++] = strdup(Buffer);
    if (psp_help_size >= MAX_HELP_LINE) break;
  }
  fclose(FileDesc);
}

static void 
psp_display_screen_help( int blit_bg )
{
  char buffer[512];

  int help_line = 0;
  int index     = 0;

  int x       = 0;
  int y       = 0;
  int y_step  = 0;

  if (blit_bg) {
    psp_sdl_blit_help();
  }

  x      = psp_sdl_get_ttf_width();
  y      = 0;
  y_step = psp_sdl_get_ttf_height();

  help_line = psp_help_current;
  index     = 0;
  
  while ((index < psp_help_line_by_page) && (help_line < psp_help_size))  {
    strcpy(buffer, psp_help[help_line]);
    string_fill_with_space(buffer, psp_help_char_by_line);
    psp_sdl_print_ttf(x, y, buffer, PSP_MENU_SEL_COLOR);
    y += y_step;
    index++;
    help_line++;
  }

  if (index != psp_help_line_by_page) {
    buffer[0]=0;
    string_fill_with_space(buffer, psp_help_char_by_line);
    while (index < psp_help_line_by_page) {
      psp_sdl_print_ttf(x, y, buffer, PSP_MENU_SEL_COLOR);
      y += y_step;
      index++;
    }
  }
}

int 
psp_help_menu()
{
  gp2xCtrlData c;
  long        new_pad;
  long        old_pad;
  int         last_time;
  int         end_menu;
  int         blit_bg;

  psp_kbd_wait_no_button();

  psp_initialize_help();

  old_pad   = 0;
  last_time = 0;
  end_menu  = 0;
  blit_bg   = 1;

  while (! end_menu)
  {
    psp_display_screen_help( blit_bg );
    psp_sdl_flip();

    while (1)
    {
      gp2xCtrlPeekBufferPositive(&c, 1);
      new_pad = c.Buttons;
# if defined(WIZ_MODE)
      if (c.Mouse) break;
# endif

      if ((old_pad != new_pad) || ((c.TimeStamp - last_time) > PSP_MENU_MIN_TIME)) {
        last_time = c.TimeStamp;
        old_pad = new_pad;
        break;
      }
    }

# if defined(WIZ_MODE)
    if (c.My > (GP2X_SDL_SCREEN_HEIGHT - 20)) {
      if (c.Mouse & GP2X_MOUSE_BTN_PRESS) new_pad = GP2X_CTRL_DOWN;
    } else
    if (c.My < 20) {
      if (c.Mouse & GP2X_MOUSE_BTN_PRESS) new_pad = GP2X_CTRL_UP;
    } else
    if (c.Mouse & (GP2X_MOUSE_GST_LEFT|GP2X_MOUSE_GST_UP)) new_pad = GP2X_CTRL_LEFT;
    else
    if (c.Mouse & (GP2X_MOUSE_GST_RIGHT|GP2X_MOUSE_GST_DOWN)) new_pad = GP2X_CTRL_RIGHT;
# endif

    if (new_pad) {
      blit_bg = 1;
    }

    if(new_pad & GP2X_CTRL_SELECT) {
      /* Back to Main menu */
      end_menu = 1;
    } else
    if(new_pad & GP2X_CTRL_UP) {
      if (psp_help_current > 0) psp_help_current--;
    } else
    if(new_pad & GP2X_CTRL_DOWN) {
      if ((psp_help_current + 1) < psp_help_size) psp_help_current++;
    } else
    if(new_pad & GP2X_CTRL_LEFT) {
      if (psp_help_current > psp_help_line_by_page) psp_help_current -= psp_help_line_by_page;
      else                                      psp_help_current  = 0;
    } else
    if(new_pad & GP2X_CTRL_RIGHT) {
      if ((psp_help_current + psp_help_line_by_page + 1) < psp_help_size) psp_help_current += psp_help_line_by_page;
    }
  }
 
  psp_kbd_wait_no_button();

  return 1;
}

