/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_ttf.h>
#include <png.h>

#include "cap32.h"
#include "video.h"
#include "psp_sdl.h"
#include "psp_danzeff.h"

  extern unsigned char psp_btm_6x10[];
  extern unsigned char psp_btm_8x8[];
  extern unsigned char psp_btm_8x10[];
  extern unsigned char psp_btm_8x12[];
  extern unsigned char psp_btm_8x14[];
  extern unsigned char psp_btm_8x16[];

  unsigned char *psp_btm        = NULL;
  int            psp_btm_width  = 8; 
  int            psp_btm_height = 8; 

  TTF_Font      *psp_ttf        = NULL;
  int            psp_ttf_width  = 8; 
  int            psp_ttf_height = 8; 

  SDL_Surface *back_surface;
  SDL_Surface *back2_surface;

  SDL_Surface *background_surface = NULL;
  SDL_Surface *blit_surface;
  SDL_Surface *help_surface;
  SDL_Surface *splash_surface;
  SDL_Surface *thumb_surface;
  SDL_Surface *save_surface;

static void
psp_sdl_init_ttf()
{
  TTF_Init();
}

static void
psp_sdl_destroy_ttf()
{
  if (psp_ttf) {
	  TTF_CloseFont( psp_ttf ); 
	  TTF_Quit(); 
    psp_ttf = NULL;
  }
}


int
psp_sdl_get_ttf_width()
{
  return psp_ttf_width;
}

int
psp_sdl_get_ttf_height()
{
  return psp_ttf_height;
}

uint
psp_sdl_rgb(uchar R, uchar G, uchar B)
{
  return SDL_MapRGB(back_surface->format, R,G,B);
}

ushort *
psp_sdl_get_vram_addr(uint x, uint y)
{
  ushort *vram = (ushort *)back_surface->pixels;
  return vram + x + (y*PSP_LINE_SIZE);
}

void 
psp_sdl_put_char_btm(int x, int y, int color, int bgcolor, uchar c, int drawfg, int drawbg)
{
  int cx;
  int cy;
  int bmask;
  int index;

  ushort *vram = (ushort *)psp_sdl_get_vram_addr(x, y);

  index = ((ushort)c) * psp_btm_height;
  for (cy=0; cy< psp_btm_height; cy++) {
    bmask = 1 << ( psp_btm_width - 1);
    for (cx=0; cx< psp_btm_width; cx++) {
      if (psp_btm[index] & bmask) {
        if (drawfg) vram[cx + cy * PSP_LINE_SIZE] = color;
      } else {
        if (drawbg) vram[cx + cy * PSP_LINE_SIZE] = bgcolor;
      }
      bmask = bmask >> 1;
    }
    index++;
  }
}

void 
psp_sdl_print_btm(int x,int y, const char *str, int color)
{
  int index;
  int x0 = x;

  for (index = 0; str[index] != '\0'; index++) {
    psp_sdl_put_char_btm(x, y, color, 0, str[index], 1, 0);
    x += psp_btm_width;
    if (x >= (PSP_SDL_SCREEN_WIDTH - psp_btm_width)) {
      x = x0; y += psp_btm_height;
    }
    if (y >= (PSP_SDL_SCREEN_HEIGHT - psp_btm_width)) break;
  }
}

void
psp_sdl_clear_screen(int color)
{
  int x; int y;
  ushort *vram = psp_sdl_get_vram_addr(0,0);
  
  for (y = 0; y < PSP_SDL_SCREEN_HEIGHT; y++) {
    for (x = 0; x < PSP_SDL_SCREEN_WIDTH; x++) {
      vram[x + (y*PSP_LINE_SIZE)] = color;
    }
  }
}

void
psp_sdl_black_screen()
{
  SDL_FillRect(back_surface,NULL,SDL_MapRGB(back_surface->format,0x0,0x0,0x0));
  SDL_Flip(back_surface);
  SDL_FillRect(back_surface,NULL,SDL_MapRGB(back_surface->format,0x0,0x0,0x0));
  SDL_Flip(back_surface);
}

void
psp_sdl_clear_blit(int color)
{
  if (blit_surface) {
    ushort *vram = (ushort *)blit_surface->pixels;
    int my_size = blit_surface->h * blit_surface->w;
    while (my_size > 0) {
      vram[--my_size] = color;
    }
  }
}

void 
psp_sdl_draw_rectangle(int x, int y, int w, int h, int border, int mode) 
{
  ushort *vram = (ushort *)psp_sdl_get_vram_addr(x, y);
  int xo, yo;
  if (mode == PSP_SDL_XOR) {
    for (xo = 0; xo < w; xo++) {
      vram[xo] ^=  border;
      vram[xo + h * PSP_LINE_SIZE] ^=  border;
    }
    for (yo = 0; yo < h; yo++) {
      vram[yo * PSP_LINE_SIZE] ^=  border;
      vram[w + yo * PSP_LINE_SIZE] ^=  border;
    }
    vram[w + h * PSP_LINE_SIZE] ^=  border;
  } else {
    for (xo = 0; xo < w; xo++) {
      vram[xo] =  border;
      vram[xo + h * PSP_LINE_SIZE] =  border;
    }
    for (yo = 0; yo < h; yo++) {
      vram[yo * PSP_LINE_SIZE] =  border;
      vram[w + yo * PSP_LINE_SIZE] =  border;
    }
    vram[w + h * PSP_LINE_SIZE] =  border;
  }
}

void 
psp_sdl_fill_rectangle(int x, int y, int w, int h, int color, int mode)
{
  ushort *vram  = (ushort *)psp_sdl_get_vram_addr(x, y);
  int xo, yo;
  if (mode == PSP_SDL_XOR) {
    for (xo = 0; xo <= w; xo++) {
      for (yo = 0; yo <= h; yo++) {
        if ( ((xo == 0) && ((yo == 0) || (yo == h))) ||
             ((xo == w) && ((yo == 0) || (yo == h))) ) {
          /* Skip corner */
        } else {
          vram[xo + yo * PSP_LINE_SIZE] ^=  color;
        }
      }
    }
  } else {
    for (xo = 0; xo <= w; xo++) {
      for (yo = 0; yo <= h; yo++) {
        vram[xo + yo * PSP_LINE_SIZE] =  color;
      }
    }
  }
}

static int
psp_sdl_get_back2_color(int x, int y)
{
  uchar *back2 = (uchar *)back2_surface->pixels;
  int bytes_per_pixels = back2_surface->format->BytesPerPixel;
  int pitch            = back2_surface->pitch;
  Uint8 r = back2[0 + (y * pitch) + (x * bytes_per_pixels)];
  Uint8 g = back2[1 + (y * pitch) + (x * bytes_per_pixels)];
  Uint8 b = back2[2 + (y * pitch) + (x * bytes_per_pixels)];
	int color = psp_sdl_rgb(r, g, b);

  return color;
}

void 
psp_sdl_back2_rectangle(int x, int y, int w, int h)
{
  if (! back2_surface) {
    psp_sdl_fill_rectangle(x, y, w, h, 0x0, 0);
    return;
  }

  ushort *vram  = (ushort *)psp_sdl_get_vram_addr(x, y);

  int xo, yo;
  for (xo = 0; xo <= w; xo++) {
    for (yo = 0; yo <= h; yo++) {
      vram[xo + yo * PSP_LINE_SIZE] = psp_sdl_get_back2_color(x + xo, y + yo);
    }
  }
}

static void 
psp_sdl_back2_put_char_btm(int x, int y, int color, uchar c)
{
  int cx;
  int cy;
  int bmask;
  int index;

  if (! back2_surface) {
    psp_sdl_put_char_btm(x, y, color, 0x0, c, 1, 1);
    return;
  }

  ushort *vram  = (ushort *)psp_sdl_get_vram_addr(x, y);

  index = ((ushort)c) * psp_btm_height;
  for (cy=0; cy< psp_btm_height; cy++) {
    bmask = 1 << ( psp_btm_width - 1);
    for (cx=0; cx< psp_btm_width; cx++) {
      if (psp_btm[index] & bmask) {
        vram[cx + cy * PSP_LINE_SIZE] = color;
      } else {
        vram[cx + cy * PSP_LINE_SIZE] = psp_sdl_get_back2_color(x + cx, y + cy);
      }
      bmask = bmask >> 1;
    }
    index++;
  }
}

unsigned char
psp_convert_utf8_to_iso_8859_1(unsigned char c1, unsigned char c2)
{
  unsigned char res = 0;
  if (c1 == 0xc2) res = c2;
  else
  if (c1 == 0xc3) res = c2 | 0x40;
  return res;
}


void 
psp_sdl_fill_print_btm(int x,int y,const char *str, int color, int bgcolor)
{
  int index;
  int x0 = x;

  for (index = 0; str[index] != '\0'; index++) {
    uchar c = str[index];
    if ((c == 0xc2) || (c == 0xc3)) {
      uchar new_c = psp_convert_utf8_to_iso_8859_1(c, str[index+1]);
      if (new_c) { c = new_c; index++; }
    }
    psp_sdl_put_char_btm(x, y, color, bgcolor, c, 1, 1);
    x += psp_btm_width;
    if (x >= (PSP_SDL_SCREEN_WIDTH - psp_btm_width)) {
      x = x0; y++;
    }
    if (y >= (PSP_SDL_SCREEN_HEIGHT - psp_btm_width)) break;
  }
}

void
psp_sdl_back2_print_btm(int x,int y,const char *str, int color)
{
  int index;
  int x0 = x;

  for (index = 0; str[index] != '\0'; index++) {
    uchar c = str[index];
    if ((c == 0xc2) || (c == 0xc3)) {
      uchar new_c = psp_convert_utf8_to_iso_8859_1(c, str[index+1]);
      if (new_c) { c = new_c; index++; }
    }
    psp_sdl_back2_put_char_btm(x, y, color, c);
    x += psp_btm_width;
    if (x >= (PSP_SDL_SCREEN_WIDTH - psp_btm_width)) {
      x = x0; y++;
    }
    if (y >= (PSP_SDL_SCREEN_HEIGHT - psp_btm_width)) break;
  }
}

void 
psp_sdl_put_char_ttf(int x, int y, int color, int bgcolor, uchar c, int drawfg, int drawbg)
{
  uchar str[2];

  str[0] = c;
  str[1] = 0;

  SDL_Surface* ttf_surface = 0;
  SDL_Color    ttf_color = { 0, 0, 0 };
  Uint32       ttf_pixel = 0;

  if (drawbg)  ttf_pixel = bgcolor;
  else         ttf_pixel = color;

  SDL_GetRGB( ttf_pixel, back_surface->format, &ttf_color.r, &ttf_color.g, &ttf_color.b );
  ttf_surface = TTF_RenderText_Blended( psp_ttf, str, ttf_color );

  if (ttf_surface) {

    if (drawbg) {
      psp_sdl_fill_rectangle( x, y, ttf_surface->w, ttf_surface->h, bgcolor, 1 );
    } else {
      psp_sdl_back2_rectangle( x, y, ttf_surface->w, ttf_surface->h );
    }

    SDL_Rect dst;
    dst.x = x; dst.y = y; dst.w = 0; dst.h = 0;
    SDL_BlitSurface( ttf_surface, NULL, back_surface, &dst );
    SDL_FreeSurface( ttf_surface );
  }
}

void
psp_sdl_print_ttf(int x,int y,const char *str, int color)
{
  SDL_Surface* ttf_surface = 0;
  SDL_Color    ttf_color = { 0, 0, 0 };
  Uint32       ttf_pixel = color;
  SDL_GetRGB( ttf_pixel, back_surface->format, &ttf_color.r, &ttf_color.g, &ttf_color.b );

  ttf_surface = TTF_RenderText_Blended( psp_ttf, str, ttf_color );

  if (ttf_surface) {
    psp_sdl_back2_rectangle( x, y, ttf_surface->w, ttf_surface->h );
    SDL_Rect dst;
    dst.x = x; dst.y = y; dst.w = 0; dst.h = 0;
    SDL_BlitSurface( ttf_surface, NULL, back_surface, &dst );
    SDL_FreeSurface( ttf_surface );
  }
}

void
psp_sdl_lock(void)
{
  SDL_LockSurface(back_surface);
}

void
psp_sdl_load_background()
{
  background_surface = IMG_Load("./background.png");
  thumb_surface = IMG_Load("./thumb.png");
}

void
psp_sdl_blit_background()
{
  static int first = 1;

  if (first && (background_surface == NULL)) {
    psp_sdl_load_background();
    first = 0;
  }
  back2_surface = background_surface;

  if (back2_surface != NULL) {
	  SDL_BlitSurface(back2_surface, NULL, back_surface, NULL);
  } else {
    psp_sdl_clear_screen(psp_sdl_rgb(0x00, 0x00, 0x00));
  }
}

void
psp_sdl_blit_thumb(int dst_x, int dst_y, SDL_Surface* thumb_surface)
{
  SDL_Rect dstRect;
  dstRect.x = dst_x;
  dstRect.y = dst_y;
  dstRect.w = thumb_surface->w;
  dstRect.h = thumb_surface->h;
  SDL_BlitSurface(thumb_surface, NULL, back_surface, &dstRect);
}

void
psp_sdl_blit_splash()
{
  if (! splash_surface) {
    splash_surface = IMG_Load("./splash.png");
  }
	SDL_BlitSurface(splash_surface, NULL, back_surface, NULL);
}

void
psp_sdl_blit_help()
{
  if (! help_surface) {
    help_surface = IMG_Load("./help.png");
  }
  back2_surface = help_surface;
	SDL_BlitSurface(back2_surface, NULL, back_surface, NULL);
}

void
psp_sdl_display_splash()
{
  int index = 0;
  char buffer[128];
  gp2xCtrlData c;

  sprintf(buffer, "v%s", CAP32_VERSION);

  int x = (PSP_SDL_SCREEN_WIDTH - 10 - (strlen(buffer) * psp_sdl_get_ttf_width()));
  int y = PSP_SDL_SCREEN_HEIGHT - 10 - psp_sdl_get_ttf_height();
  int col = psp_sdl_rgb(0xa0, 0xa0, 0xa0);


  psp_sdl_blit_splash();
  psp_sdl_print_ttf(x, y, buffer, col);
  psp_sdl_flip();

  while (index < 10) {
    SDL_Delay( 150 );
    gp2xCtrlPeekBufferPositive(&c, 1);
    if (c.Buttons & (GP2X_CTRL_START|GP2X_CTRL_CROSS)) break;
    index++;
  }
}

void
psp_sdl_unlock(void)
{
  SDL_UnlockSurface(back_surface);
}

void
psp_sdl_flip(void)
{
  SDL_Flip(back_surface);
}

#define  systemRedShift      (back_surface->format->Rshift)
#define  systemGreenShift    (back_surface->format->Gshift)
#define  systemBlueShift     (back_surface->format->Bshift)
#define  systemRedMask       (back_surface->format->Rmask)
#define  systemGreenMask     (back_surface->format->Gmask)
#define  systemBlueMask      (back_surface->format->Bmask)

int
psp_sdl_save_png(SDL_Surface* my_surface, char* filename)
{
  int w = my_surface->w;
  int h = my_surface->h;
  u8* pix = (u8*)my_surface->pixels;
  u8 writeBuffer[800 * 3];

  FILE *fp = fopen(filename,"wb");

  if(!fp) return 0;
  
  png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
                                                NULL,
                                                NULL,
                                                NULL);
  if(!png_ptr) {
    fclose(fp);
    return 0;
  }

  png_infop info_ptr = png_create_info_struct(png_ptr);

  if(!info_ptr) {
    png_destroy_write_struct(&png_ptr,NULL);
    fclose(fp);
    return 0;
  }

  png_init_io(png_ptr,fp);

  png_set_IHDR(png_ptr,
               info_ptr,
               w,
               h,
               8,
               PNG_COLOR_TYPE_RGB,
               PNG_INTERLACE_NONE,
               PNG_COMPRESSION_TYPE_DEFAULT,
               PNG_FILTER_TYPE_DEFAULT);

  png_write_info(png_ptr,info_ptr);

  u8 *b = writeBuffer;

  int sizeX = w;
  int sizeY = h;
  int y;
  int x;

  u16 *p = (u16 *)pix;
  for(y = 0; y < sizeY; y++) {
     for(x = 0; x < sizeX; x++) {
      u16 v = p[x];

      *b++ = ((v & systemRedMask  ) >> systemRedShift  ) << 3; // R
      *b++ = ((v & systemGreenMask) >> systemGreenShift) << 2; // G 
      *b++ = ((v & systemBlueMask ) >> systemBlueShift ) << 3; // B
    }
    p += my_surface->pitch / 2;
    png_write_row(png_ptr,writeBuffer);
     
    b = writeBuffer;
  }

  png_write_end(png_ptr, info_ptr);

  png_destroy_write_struct(&png_ptr, &info_ptr);

  fclose(fp);
  return 1;
}

int
psp_sdl_load_png(SDL_Surface* my_surface, char* filename)
{
  int w = my_surface->w;
  int h = my_surface->h;
  int pitch = my_surface->pitch / 2;
  u16* pix = (u16*)my_surface->pixels;

  FILE *fp = fopen(filename,"rb");
  if (!fp) return 0;

  const size_t nSigSize = 8;
  u8 signature[nSigSize];
  if (fread(signature, sizeof(u8), nSigSize, fp) != nSigSize) {
    fclose(fp);
    return 0;
  }

  if (!png_check_sig(signature, nSigSize)) {
    fclose(fp);
    return 0;
  }

  png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
                                               NULL,
                                               NULL,
                                               NULL);
  if(!png_ptr) {
    fclose(fp);
    return 0;
  }

  png_infop info_ptr = png_create_info_struct(png_ptr);

  if(!info_ptr) {
    png_destroy_read_struct(&png_ptr, NULL, NULL);
    fclose(fp);
    return 0;
  }

  png_init_io(png_ptr, fp);
  png_set_sig_bytes(png_ptr, nSigSize);
  png_read_png(png_ptr, info_ptr,
    PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_PACKING |
    PNG_TRANSFORM_EXPAND | PNG_TRANSFORM_BGR , NULL);

  png_uint_32 width = info_ptr->width;
  png_uint_32 height = info_ptr->height;
  int color_type = info_ptr->color_type;

  if ((width  > w) ||
      (height > h)) {
    png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
    fclose(fp);
    return 0;
  }

  png_byte **pRowTable = info_ptr->row_pointers;
  unsigned int x, y;
  u8 r, g, b;

  for (y=0; y<height; y++) {
    png_byte *pRow = pRowTable[y];
    for (x=0; x<width; x++) {
      switch(color_type) {
        case PNG_COLOR_TYPE_GRAY:
          r = g = b = *pRow++;
          break;
        case PNG_COLOR_TYPE_GRAY_ALPHA:
          r = g = b = pRow[0];
          pRow += 2;
          break;
        case PNG_COLOR_TYPE_RGB:
          b = pRow[0];
          g = pRow[1];
          r = pRow[2];
          pRow += 3;
          break;
        case PNG_COLOR_TYPE_RGB_ALPHA:
          b = pRow[0];
          g = pRow[1];
          r = pRow[2];
          pRow += 4;
          break;
        default:
          r = g = b = 0;
          break;
      }

      u16 v = (((r >> 3) << systemRedShift  ) & systemRedMask) |
              (((g >> 2) << systemGreenShift) & systemGreenMask) |
              (((b >> 3) << systemBlueShift ) & systemBlueMask);
      *pix++= v;
    }
    pix += pitch - width;
  }
  png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
  fclose(fp);

  return 1;
}

int
psp_sdl_save_screenshot_png(char *filename)
{
  return psp_sdl_save_png(back_surface, filename);
}

int
psp_sdl_save_thumb_png(SDL_Surface* my_surface, char* filename)
{
  /* First dump blit_surface to my_surface */
  int x;
  int y;
  u16* src_pixel = (u16*)blit_surface->pixels;
  u16* dst_pixel = (u16*)my_surface->pixels;
  u16* scan_src_pixel = 0;

  for (y = 0; y < SNAP_HEIGHT; y++) {
    scan_src_pixel  = src_pixel + (blit_surface->w * y);
    scan_src_pixel += (blit_surface->w - SNAP_WIDTH) / 2;
    for (x = 0; x < SNAP_WIDTH; x++) {
      *dst_pixel++ = scan_src_pixel[x];
    }
  }
  /* Then save thumb in file */
  return psp_sdl_save_png(my_surface, filename);
}

int
psp_sdl_load_thumb_png(SDL_Surface* my_surface, char* filename)
{ 
  return psp_sdl_load_png( my_surface, filename);
}

void
psp_sdl_save_screenshot(void)
{
  char TmpFileName[MAX_PATH];

  sprintf(TmpFileName,"%s/screenshot_%d.png", CPC.psp_screenshot_path, CPC.psp_screenshot_id++);
  if (CPC.psp_screenshot_id >= 10) CPC.psp_screenshot_id = 0;
  psp_sdl_save_screenshot_png(TmpFileName);
}

void
psp_sdl_select_font_tt( int size )
{
  if (psp_ttf) {
    TTF_CloseFont( psp_ttf );
    psp_ttf = NULL;
  }
  psp_ttf = TTF_OpenFont( CPC.font_filename, size );
  TTF_SizeText( psp_ttf, "W", &psp_ttf_width, &psp_ttf_height );
}

void
psp_sdl_select_font_8x8()
{
  psp_btm        = psp_btm_8x8;
  psp_btm_width  = 8;
  psp_btm_height = 8;
}

void
psp_sdl_select_font_8x10()
{
  psp_btm        = psp_btm_8x10;
  psp_btm_width  = 8;
  psp_btm_height = 10;
}

void
psp_sdl_select_font_8x12()
{
  psp_btm        = psp_btm_8x12;
  psp_btm_width  = 8;
  psp_btm_height = 12;
}

void
psp_sdl_select_font_8x14()
{
  psp_btm        = psp_btm_8x14;
  psp_btm_width  = 8;
  psp_btm_height = 14;
}

void
psp_sdl_select_font_8x16()
{
  psp_btm        = psp_btm_8x16;
  psp_btm_width  = 8;
  psp_btm_height = 16;
}

void
psp_sdl_select_font_6x10()
{
  psp_btm        = psp_btm_6x10;
  psp_btm_width  = 6;
  psp_btm_height = 10;
}



int
psp_sdl_init(void)
{
  psp_sdl_default_config();
  psp_sdl_load_config();

  if (SDL_Init(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK|SDL_INIT_AUDIO) < 0) {
     return 0;
  }
  SDL_JoystickEventState(SDL_ENABLE);
  SDL_JoystickOpen(0);

  if (SDL_InitSubSystem(SDL_INIT_VIDEO) < 0) {
    return 0;
  }

# ifdef PND_MODE
  back_surface=SDL_SetVideoMode(PSP_SDL_SCREEN_WIDTH,PSP_SDL_SCREEN_HEIGHT, 16 , SDL_SWSURFACE|SDL_FULLSCREEN);
# else
  back_surface=SDL_SetVideoMode(PSP_SDL_SCREEN_WIDTH,PSP_SDL_SCREEN_HEIGHT, 16 , SDL_SWSURFACE);
# endif

  if ( !back_surface) {
    return 0;
  }

  psp_sdl_init_ttf();
  psp_sdl_select_font_tt( CPC.font_size );
  psp_sdl_select_font_8x14();

  blit_surface = SDL_CreateRGBSurface(SDL_SWSURFACE, 
    CPC_SCR_WIDTH, CPC_SCR_HEIGHT,
    back_surface->format->BitsPerPixel,
    back_surface->format->Rmask,
    back_surface->format->Gmask,
    back_surface->format->Bmask, 0);
  SDL_ShowCursor(SDL_DISABLE);

  psp_sdl_display_splash();

  /* Danzeff Keyboard */
  danzeff_load();
  danzeff_set_screen(back_surface);

  /* Create surface for save state */
  int Index = 0;
  for (Index = 0; Index < CPC_MAX_SAVE_STATE; Index++) {
    CPC.cpc_save_state[Index].surface = 
       SDL_CreateRGBSurface(SDL_SWSURFACE, 
                            SNAP_WIDTH, SNAP_HEIGHT,
                            back_surface->format->BitsPerPixel,
                            back_surface->format->Rmask,
                            back_surface->format->Gmask,
                            back_surface->format->Bmask, 0);
  }
  save_surface = SDL_CreateRGBSurface(SDL_SWSURFACE, 
                            SNAP_WIDTH, SNAP_HEIGHT,
                            back_surface->format->BitsPerPixel,
                            back_surface->format->Rmask,
                            back_surface->format->Gmask,
                            back_surface->format->Bmask, 0);

  return 1;
}

void
psp_sdl_exit(int status)
{
  psp_sdl_save_config();
  psp_sdl_destroy_ttf();

  SDL_Quit();
  exit(status);
}

void
psp_sdl_set_palette(SDL_Color* c) 
{
	SDL_SetPalette(back_surface, SDL_LOGPAL | SDL_PHYSPAL, c, 0, 32); 
}

void
psp_sdl_default_config()
{
  getcwd(CPC.snap_directory, sizeof(CPC.snap_directory));
  strcpy(CPC.disk_directory, CPC.snap_directory);
  strcat(CPC.snap_directory , "/snap/");
  strcat(CPC.disk_directory, "/disk/");
  strcpy(CPC.font_filename, "font.ttf");
  CPC.font_size = 16;
}

int
psp_sdl_load_config()
{
  char  Buffer[512];
  char *Scan;
  unsigned int Value;
  FILE* FileDesc;

  FileDesc = fopen( "global.cfg", "r");
  if (FileDesc == (FILE *)0 ) return 0;

  while (fgets(Buffer,512, FileDesc) != (char *)0) {

    Scan = strchr(Buffer,'\n');
    if (Scan) *Scan = '\0';
    /* For this #@$% of windows ! */
    Scan = strchr(Buffer,'\r');
    if (Scan) *Scan = '\0';
    if (Buffer[0] == '#') continue;

    Scan = strchr(Buffer,'=');
    if (! Scan) continue;

    *Scan = '\0';
    if (!strcasecmp(Buffer,"snap_directory")) {
      strcpy(CPC.snap_directory, Scan+1);
    } else
    if (!strcasecmp(Buffer,"disk_directory")) {
      strcpy(CPC.disk_directory, Scan+1);
    } else
    if (!strcasecmp(Buffer,"font_filename")) {
      strcpy(CPC.font_filename, Scan+1);
    } else
    if (!strcasecmp(Buffer,"font_size")) {
      CPC.font_size = atoi(Scan+1);
    }
  }

  fclose(FileDesc);

  return 0;
}

int
psp_sdl_save_config()
{
  FILE* FileDesc;
  int   error = 0;

  FileDesc = fopen("global.cfg", "w");
  if (FileDesc != (FILE *)0 ) {

    fprintf(FileDesc, "snap_directory=%s\n", CPC.snap_directory);
    fprintf(FileDesc, "disk_directory=%s\n", CPC.disk_directory);
    fprintf(FileDesc, "font_filename=%s\n" , CPC.font_filename);
    fprintf(FileDesc, "font_size=%d\n"     , CPC.font_size);

    fclose(FileDesc);

  } else {
    error = 1;
  }

  return error;
}


